#!/usr/bin/env python
#^#
#
# <description>
#
#
# Copyright (C) 2006, 2007 Oracle.  All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 021110-1307, USA.
#$#

import signal
import re
import os
import logging
from os.path import dirname, basename, join, isdir, isfile, exists
from glob import glob
from OVSServerProxy import ServerProxy
from OVSConstants import HA_TOUCH_PREFIX, HA_NFS_HEARTBEAT_FILE, OVS_BASE, PID_DIR
from OVSConstants import DLM_DOMAIN, AGENT_GUARD_PROCESS, AGENT_DIR, DLM
from OVSCommons import fmt_err_msg, exposed, parse_result, marshall_string, unmarshall_string, run_cmd, run_cmd_bg, old_run_cmd, parse_shared_fs
from OVSCDS import CDS
from OVSXAPIUtil import get_running_vms

@exposed
def ha_check_exist(path):
    logger = logging.getLogger("ovs.query")
    try:
        _dirname = dirname(path)
        filename = "%s%s" % (HA_TOUCH_PREFIX, basename(path))
        touch_file = join(_dirname, filename)
        if not isfile(touch_file):
            raise Exception("file touch path('%s') does not exists!" % touch_file)
        logger.info("ha_check_exist: success. file touch path('%s')." % touch_file)
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_check_exist: failed. => %s" % errmsg)
        return "failed:%s" % errmsg


def _find_ocfs2_dev(uuid, dev_peer):
    # dev_peer: "dev/sdb1" -> ptn: "/dev/sd*1"
    v = list(dev_peer)
    idx = -1
    while v[idx] in "0123456789":
        idx -= 1
    v[idx] = "*"
    ptn = "".join(v)
    devs = glob(ptn)

    #find dev with same uuid
    tgt_dev = []
    for dev in devs:
        try:
            _uuid = old_run_cmd("debugfs.ocfs2 %s <<<stats 2>/dev/null | awk '/UUID/ {print $2;}'" % dev).splitlines()[-1]
            if _uuid == uuid:
                tgt_dev.append(dev)
        except:
            pass

    return tgt_dev


def get_ocfs2_nodes():
    n_path = "/sys/kernel/config/cluster/ocfs2/node"
    if exists(n_path):
        return os.listdir(n_path)
    else:
        return []

def check_ocfs2_nodes(master_nodes):
    m_nodes = set(master_nodes)
    s_nodes = set(get_ocfs2_nodes())
    return s_nodes.issubset(m_nodes)

@exposed
def check_shared_storage_mount(marshalled_shared_fs_line, marshalled_shared_repositories, shared_fs_extra, master_nodes):
    logger = logging.getLogger("ovs.operation")
    try:
        #base shared storage
        shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()

        master_shared_fs_line = unmarshall_string(marshalled_shared_fs_line)
        master_shared_fs, master_mount_point, master_shared_fs_type, master_mount_option = master_shared_fs_line.split()[0:4]

        if master_shared_fs_type == "ocfs2":
            if not check_ocfs2_nodes(master_nodes):
                raise Exception("Please clean up OCFS2 misconfigurations.")

        if shared_fs and shared_fs_type == "ocfs2" and master_shared_fs_type == "ocfs2":
            #already mounted and is ocfs2 partition mounted by device name
            uuid = parse_result("success:%s" % shared_fs_extra)["uuid"]
            tgt_dev = _find_ocfs2_dev(uuid, master_shared_fs)
            if shared_fs not in tgt_dev:
                raise Exception("%s already mounted but is not the same as master node." % OVS_BASE)
        elif shared_fs and (master_shared_fs_type != shared_fs_type or master_shared_fs != shared_fs):
            #already mounted, but not the same as master node
                raise Exception("%s already mounted but is not the same as master node." % OVS_BASE)
        elif not shared_fs:
            #local filesytem, try to mount the shared storage
            if master_shared_fs_type == "ocfs2" and master_shared_fs.startswith("/dev"):
                uuid = parse_result("success:%s" % shared_fs_extra)["uuid"]
                tgt_dev = _find_ocfs2_dev(uuid, master_shared_fs)
                if not tgt_dev:
                    raise Exception("cannot find the equivalent filesystem as master partition with uuid='%s'" % uuid)

        logger.info("check_shared_storage_mount: success")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("check_shared_storage_mount: error. =>%s" % errmsg)
        return "failed:%s" % errmsg


@exposed
def ha_mount_shared_storage(marshalled_shared_fs_line, marshalled_shared_repositories, shared_fs_extra=""):
    logger = logging.getLogger("ovs.operation")
    try:
        #base shared storage
        shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()

        master_shared_fs_line = unmarshall_string(marshalled_shared_fs_line)
        master_shared_fs, master_mount_point, master_shared_fs_type, master_mount_option = master_shared_fs_line.split()[0:4]

        if shared_fs and shared_fs_type == "ocfs2" and master_shared_fs_type == "ocfs2":
            #already mounted and is ocfs2 partition mounted by device name
            uuid = parse_result("success:%s" % shared_fs_extra)["uuid"]
            tgt_dev = _find_ocfs2_dev(uuid, master_shared_fs)
            if shared_fs not in tgt_dev:
                raise Exception("%s already mounted but is not the same as master node." % OVS_BASE)
        elif shared_fs and (master_shared_fs_type != shared_fs_type or master_shared_fs != shared_fs):
            #already mounted, but not the same as master node
                raise Exception("%s already mounted but is not the same as master node." % OVS_BASE)
        elif not shared_fs:
            #local filesytem, try to mount the shared storage
            if master_shared_fs_type == "ocfs2" and master_shared_fs.startswith("/dev"):
                uuid = parse_result("success:%s" % shared_fs_extra)["uuid"]
                tgt_dev = _find_ocfs2_dev(uuid, master_shared_fs)
                if tgt_dev:
                    shared_fs = tgt_dev[0]
                else:
                    raise Exception("cannot find the equivalent filesystem as master partition with uuid='%s'" % uuid)
            else:
                shared_fs = master_shared_fs
            #cmd = "mount -t %s -o %s %s %s" % (master_shared_fs_type, master_mount_option, shared_fs, master_mount_point)
            cmdlist = ["mount", "-t", master_shared_fs_type, "-o", master_mount_option, shared_fs, master_mount_point]
            run_cmd(args=cmdlist)

        #extended shared storage
        shared_repositories = unmarshall_string(marshalled_shared_repositories)
        f = open("/etc/ovs/repositories", "w")
        try:
            f.write(shared_repositories)
        finally:
            f.close()
        #cmd = "service ovsrepositories start"
        cmdlist = ["service", "ovsrepositories", "start"]
        run_cmd(args=cmdlist)

        logger.info("ha_mount_shared_storage: success")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_mount_shared_storage: error. =>%s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_setup_ocfs2_cluster(marshalled_cluster_config, shared_fs_type="ocfs2"):
    """
    *) reconfigure the o2cb service
    *) start the heartbeating
        /OVS is ocfs2, mount the partition
        /OVS is nfs, start heartbeat against the dmnfs heartbeat device
    """
    logger = logging.getLogger("ovs.operation")
    try:
        #configure & start cluster service
        if not isdir("/etc/ocfs2"):
            os.makedirs("/etc/ocfs2")
        f = open("/etc/ocfs2/cluster.conf", "w")
        try:
            f.write(unmarshall_string(marshalled_cluster_config))
        finally:
            f.close()

        logger.info("ha_setup_ocfs2_cluster: trying to reconfigure the cluster...")
        cmd = join(dirname(__file__), "utils/config_o2cb.sh")
        cmdlist = [cmd]
        run_cmd(args=cmdlist)

        #start heartbeating
        if shared_fs_type == "ocfs2":
            #mount will let the server's start hearting, see ha_sync_shared_storage_mount()
            pass
        elif shared_fs_type == "nfs":
            #cmd = "/usr/lib/ovs/ovs-cluster-check --alter-fstab"
            cmdlist = ["/usr/lib/ovs/ovs-cluster-check", "--alter-fstab"]
            run_cmd(args=cmdlist)
        else:
            raise Exception("filesystem type '%s' not in [ocfs2, nfs]." % shared_fs_type)

        logger.info("ha_setup_ocfs2_cluster: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_setup_ocfs2_cluster: error. =>%s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_check_cluster_heartbeat():
    """
    check wheter the ocfs2 cluster heartbeat is alive
    """
    logger = logging.getLogger("ovs.query")
    try:
        rs = "failed"
        shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()
        if shared_fs_type == "ocfs2":
            #cmd = "ocfs2_hb_ctl -I -d %s" % shared_fs
            cmdlist = ["ocfs2_hb_ctl", "-I", "-d", shared_fs]
            l = run_cmd(args=cmdlist).splitlines()
            if l[-1].split()[1] != "0":
                rs = "success"
        elif shared_fs_type == "nfs":
            #cmd = "dmsetup table --target=nfs"
            cmdlist = ["dmsetup", "table", "--target=nfs"]
            l = run_cmd(args=cmdlist).splitlines()
            for line in l:
                parts = line.split()
                if len(parts) > 4 and parts[4] == HA_NFS_HEARTBEAT_FILE:
                    dev = parts[0].split(":")[0]
                    #cmd = "ocfs2_hb_ctl -I -d /dev/mapper/%s" % dev
                    cmdlist = ["ocfs2_hb_ctl", "-I", "-d", "/dev/mapper/%s" % dev]
                    _l = run_cmd(args=cmdlist).splitlines()
                    if _l[-1].split()[1] != "0":
                        rs = "success"
                        break
        else:
            raise Exception("only [ocfs2, nfs] shared filesystem is supported.")

        logger.info("ha_check_cluster_heartbeat: shared_fs='%s' %s." % (shared_fs, rs))
        return rs
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_check_cluster_heartbeat: error. %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def ha_join_dlm_domain(domain_name="ovm"):
    logger = logging.getLogger("ovs.operation")
    try:
        rs = ha_check_cluster_heartbeat()
        if not rs.startswith("success"):
            raise Exception("server is not heartbeating!")

        _dir = join(DLM, basename(domain_name))
        if not isdir(_dir):
            #find heartbeat device
            hb_dev = ""
            shared_fs, shared_fs_type, shared_fs_line = parse_shared_fs()
            if shared_fs_type == "ocfs2":
                hb_dev = shared_fs
            elif shared_fs_type == "nfs":
                #cmd = "dmsetup table --target=nfs"
                cmdlist = ["dmsetup", "table", "--target=nfs"]
                l = run_cmd(args=cmdlist).splitlines()
                for line in l:
                    parts = line.split()
                    if len(parts) > 4 and parts[4] == HA_NFS_HEARTBEAT_FILE:
                        hb_dev = join("/dev/mapper", parts[0].split(":")[0])
                        break
            else:
                raise Exception("currently only [ocfs2, nfs] shared filesystems are supported.")

            #add heartbeat reference
            #cmd = "ocfs2_hb_ctl -S -d %s" % hb_dev
            cmdlist = ["ocfs2_hb_ctl", "-S", "-d", hb_dev]
            run_cmd(args=cmdlist)

            #join domain
            os.mkdir(_dir)
            logger.info("ha_join_dlm_domain: success.")

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_join_dlm_domain: domain('%s') error. => %s" % (domain_name, errmsg))
        return "failed:%s" % errmsg


def ha_register_lock(uuid, lock_name):
    logger = logging.getLogger("ovs.operation")
    try:
        cds = CDS("master")
        try:
            mdb = cds.db_master
            sp = ServerProxy(mdb["master_agt_url"])
            rs = sp.register_dlm_lock(lock_name, uuid, mdb["slave_hostname"])
        finally:
            cds.close()

        if not rs.startswith("success"):
            raise Exception(rs)

        logger.info("ha_register_lock: success. uuid('%s'), lock_name('%s')" % (uuid, lock_name))
        return rs

    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_register_lock: failed. uuid('%s'), lock_name('%s') => %s" % (uuid, lock_name, errmsg))
        return "failed:%s" % errmsg


def ha_unregister_lock(lock_name):
    logger = logging.getLogger("ovs.operation")
    try:
        cds = CDS("master")
        try:
            mdb = cds.db_master
            sp = ServerProxy(mdb["master_agt_url"])
            rs = sp.unregister_dlm_lock(lock_name)
        finally:
            cds.close()

        if not rs.startswith("success"):
            raise Exception(rs)

        logger.info("ha_unregister_lock: success. lock_name('%s')" % lock_name)
        return rs

    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_unregister_lock: failed. lock_name('%s') => %s" % (lock_name, errmsg))
        return "failed:%s" % errmsg


@exposed
def ha_set_dlm_lock(lock, name=""):
    """
    @param lock uuid of the vm
    @param name name of the vm
    @return     "success" if the lock successfully or "failed" with the error message
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if not lock.strip():
            raise Exception("lock must be specified!")

        rs = ha_join_dlm_domain(DLM_DOMAIN)
        if not rs.startswith("success"):
            raise Exception(rs)

        #limit dlm lock name to 31 chars
        lock_path = join(DLM_DOMAIN, re.sub(r"[^a-zA-Z0-9]+", "", lock)[-31:])

        if exists(lock_path):
            raise Exception("lock('%s') has already been set!" % lock_path)

        try:
            fd = os.open(lock_path, os.O_CREAT | os.O_NONBLOCK)
            os.close(fd)
        except Exception, e:
            try:
                os.remove(lock_path) #get rid of the side effect
            except:
                pass
            errmsg = fmt_err_msg(e)
            raise Exception("create lock('%s') failed. %s" % (lock_path, errmsg))

        #create the EX lock
        fd = os.open(lock_path, os.O_RDWR | os.O_NONBLOCK)
        #close it, or the agent process will hold the lock
        os.close(fd)

        rs = ha_register_lock(lock, name)
        if not rs.startswith("success"):
            raise Exception(rs)

        logger.info("ha_set_dlm_lock: lock('%s') has been set for vm('%s')" % (lock, name))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_set_dlm_lock:failed. lock('%s') name('%s')=> %s" % (lock, name, errmsg))
        return "failed: %s" % errmsg

ha_accquire_lock = ha_set_dlm_lock

@exposed
def ha_release_dlm_lock(lock, name=""):
    """
    @param lock uuid of the vm
    @param name name of the vm
    @return     "success" if the lock successfully or "failed" with the error message
    """
    logger = logging.getLogger("ovs.operation")
    try:
        if not lock.strip():
            raise Exception("lock must be specified!")

        lock_path = join(DLM_DOMAIN, re.sub(r"[^a-zA-Z0-9]+", "", lock)[-31:])
        if exists(lock_path):
            os.remove(lock_path)

        rs = ha_unregister_lock(name)
        if not rs.startswith("success"):
            raise Exception(rs)

        logger.info("ha_release_dlm_lock:lock('%s') name('%s') success." % (lock, name))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_release_dlm_lock:failed. lock('%s') name('%s') => %s" % (lock, name, errmsg))
        return "failed: %s" % errmsg

ha_release_lock = ha_release_dlm_lock


@exposed
def ha_start_agent_guard():
    logger = logging.getLogger("ovs.operation")
    try:
        run_cmd_bg(AGENT_GUARD_PROCESS)

        logger.info("ha_start_agent_guard: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_start_agent_guard: error. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_stop_agent_guard():
    logger = logging.getLogger("ovs.operation")
    try:
        pid_file = "%s.pid" % basename(AGENT_GUARD_PROCESS).split(".")[0]
        pid_file = join(PID_DIR, pid_file)
        f = open(pid_file)
        try:
            pid = f.read().strip()
        finally:
            f.close()

        #release nowayout lock
        for i in range(3):
            #cmd = "kill -%d %s" % (signal.SIGUSR1, pid)
            cmdlist = ["kill", "-%d" % signal.SIGUSR1, pid]
            run_cmd(args=cmdlist)
        #ternimate the process
        #cmd = "kill -%d %s" % (signal.SIGKILL, pid)
        cmdlist = ["kill", "-%d" % signal.SIGKILL, pid]
        run_cmd(args=cmdlist)

        logger.info("ha_stop_agent_guard: success.")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_stop_agent_guard: error. => %s" % errmsg)
        return "failed:%s" % errmsg

@exposed
def ha_pull_cluster_setup():
    """
    when agent startup, pull cluster configuration from master node and set it up.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        master_agt_url = ""
        cds = CDS("master")
        try:
            master_agt_url = cds.db_master.get("master_agt_url")
        finally:
            cds.close()
        if not master_agt_url:
            logger.warning("ha_pull_cluster_setup: no master access information available!")
            return "success"

        sp = ServerProxy(master_agt_url)
        #shared storage configuration
        rs = sp.ha_get_shared_fs_conf()
        if not rs.startswith("success"):
            raise Exception(rs)
        marshalled_shared_fs_line, marshalled_shared_repositories, shared_fs_extra = rs[7:].strip().split("@@@")

        shared_fs_line = unmarshall_string(marshalled_shared_fs_line).strip()
        if not shared_fs_line:
            raise Exception("shared storage is not mounted on master server!")
        parts = shared_fs_line.split()
        shared_fs, shared_fs_type = parts[0], parts[2]

        #cluster configuration
        rs = sp.ha_get_cluster_conf()
        if not rs.startswith("success"):
            raise Exception(rs)
        marshalled_cluster_config = rs[8:] #strip "success:"

        #setup
        if shared_fs_type == "ocfs2":
            rs = ha_setup_ocfs2_cluster(marshalled_cluster_config, shared_fs_type)
            if not rs.startswith("success"):
                raise Exception(rs)

            rs = ha_mount_shared_storage(marshalled_shared_fs_line, marshalled_shared_repositories, shared_fs_extra)
            if not rs.startswith("success"):
                raise Exception(rs)

        elif shared_fs_type == "nfs":
            rs = ha_mount_shared_storage(marshalled_shared_fs_line, marshalled_shared_repositories, shared_fs_extra)
            if not rs.startswith("success"):
                raise Exception(rs)

            rs = ha_setup_ocfs2_cluster(marshalled_cluster_config, shared_fs_type)
            if not rs.startswith("success"):
                raise Exception(rs)
        else:
            raise Exception("%s not in supported shared filesystem [ocfs2, nfs]!" % shared_fs_type)

        rs = ha_start_agent_guard()
        if not rs.startswith("success"):
            raise Exception(rs)

        logger.info("ha_pull_cluster_setup:success")
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_pull_cluster_setup:failed. => %s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def ha_set_shutdown_mode(action="enter"):
    """
    1) tell the master the server is in shutdown mode
    2) create the shutdown mode flag
    """
    logger = logging.getLogger("ovs.operation")
    try:
        logger.info("ha_set_shutdown_mode: inform master agent, %s shutdown mode." % action)
        master_agt_url = ""
        cds = CDS("master")
        try:
            master_agt_url = cds.db_master.get("master_agt_url")
        finally:
            cds.close()
        if not master_agt_url:
            logger.warning("ha_set_shutdown_mode: no master access information available!")
        else:
            from OVSCommons import get_my_ips
            my_ips = get_my_ips()
            my_ips = [ip for ip in my_ips if not ip.startswith("127")]
            sp = ServerProxy(master_agt_url)
            rs = "success"
            if action == "enter":
                rs = sp.set_shutdown_mode(",".join(my_ips), True)
            elif action == "leave":
                rs = sp.set_shutdown_mode(",".join(my_ips), False)

            if not rs.startswith("success"):
                raise rs

        from OVSConstants import AGENT_SHUTDOWN_MODE_FLAG
        if action == "enter":
            logger.info("ha_set_shutdown_mode: create shutdown mode flag.")
            open(AGENT_SHUTDOWN_MODE_FLAG, "w").close()
        elif action == "leave":
            logger.info("ha_set_shutdown_mode: remove shutdown mode flag.")
            if exists(AGENT_SHUTDOWN_MODE_FLAG):
                os.remove(AGENT_SHUTDOWN_MODE_FLAG)

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_set_shutdown_mode: failed. =>%s" % errmsg)
        return "failed: %s" % errmsg

@exposed
def ha_stop_heartbeat():
    """
    xref: ha_remove_cluster_infrastructure()
    if /OVS is nfs:
        stop heartbeat against heartbeat file/device.
        utils/stop_heartbeat_nfs.sh

    if /OVS is ocfs2:
        do nothing. this function will not be called at all.
    """
    logger = logging.getLogger("ovs.operation")
    try:
        cmd = join(AGENT_DIR, "utils/stop_heartbeat_nfs.sh")
        cmdlist = [cmd]
        run_cmd(args=cmdlist)

        return "success"
    except Exception, e:
        errmsg = fmt_err_msg(e)
        logger.error("ha_stop_heartbeat: failed. => %s" % errmsg)
        return "failed:%s" % errmsg


@exposed
def ha_reset_dlm_lock(locks):
    logger = logging.getLogger("ovs.operation")
    logger.info("ha_reset_dlm_lock: start...")
    try:
        vms = get_running_vms()
        locks_name = [ k for k, v in locks ]
        vms_name = [ k for k, v in vms ]

        for k, v in vms:
            if not k in locks_name:
                ha_set_dlm_lock(v, k)

        for k, v in locks:
            if not k in vms_name:
                ha_release_dlm_lock(v, k)

        logger.info("ha_reset_dlm_lock: success. locks('%s')" % repr(locks))
        return "success"
    except Exception, e:
        errmsg = fmt_err_msg()
        logger.error("ha_reset_dlm_lock: failed. locks('%s') => %s" % (repr(locks), errmsg))
        return "failed:%s" % errmsg
